@@ -1,32 +1,40 @@
-# Copyright 2012-2015, Andrey Kislyuk and argcomplete contributors.
+# Copyright 2012-2021, Andrey Kislyuk and argcomplete contributors.
 # Licensed under the Apache License. See https://github.com/kislyuk/argcomplete for more info.
 
-from __future__ import absolute_import, division, print_function, unicode_literals
+import argparse
+import contextlib
+import os
+import sys
 
-import os, sys, argparse, contextlib
-from . import completers, my_shlex as shlex
-from .compat import USING_PYTHON2, str, sys_encoding, ensure_str, ensure_bytes
-from .completers import FilesCompleter
-from .my_argparse import IntrospectiveArgumentParser, action_is_satisfied, action_is_open, action_is_greedy
+from . import completers
+from . import my_shlex as shlex
+from .completers import FilesCompleter, SuppressCompleter
+from .my_argparse import IntrospectiveArgumentParser, action_is_greedy, action_is_open, action_is_satisfied
+from .shell_integration import shellcode  # noqa
 
 _DEBUG = "_ARC_DEBUG" in os.environ
 
 debug_stream = sys.stderr
 
+
 def debug(*args):
     if _DEBUG:
         print(file=debug_stream, *args)
 
+
 BASH_FILE_COMPLETION_FALLBACK = 79
 BASH_DIR_COMPLETION_FALLBACK = 80
 
-safe_actions = (argparse._StoreAction,
-                argparse._StoreConstAction,
-                argparse._StoreTrueAction,
-                argparse._StoreFalseAction,
-                argparse._AppendAction,
-                argparse._AppendConstAction,
-                argparse._CountAction)
+safe_actions = (
+    argparse._StoreAction,
+    argparse._StoreConstAction,
+    argparse._StoreTrueAction,
+    argparse._StoreFalseAction,
+    argparse._AppendAction,
+    argparse._AppendConstAction,
+    argparse._CountAction,
+)
+
 
 @contextlib.contextmanager
 def mute_stdout():
@@ -37,6 +45,7 @@ def mute_stdout():
     finally:
         sys.stdout = stdout
 
+
 @contextlib.contextmanager
 def mute_stderr():
     stderr = sys.stderr
@@ -47,13 +56,18 @@ def mute_stderr():
         sys.stderr.close()
         sys.stderr = stderr
 
+
 class ArgcompleteException(Exception):
     pass
 
+
 def split_line(line, point=None):
     if point is None:
         point = len(line)
-    lexer = shlex.shlex(line, posix=True, punctuation_chars=True)
+    line = line[:point]
+    lexer = shlex.shlex(line, posix=True)
+    lexer.whitespace_split = True
+    lexer.wordbreaks = os.environ.get("_ARGCOMPLETE_COMP_WORDBREAKS", "")
     words = []
 
     def split_word(word):
@@ -74,9 +88,7 @@ def split_line(line, point=None):
         # if len(prefix) > 0 and prefix[0] in lexer.quotes:
         #    prequote, prefix = prefix[0], prefix[1:]
 
-        first_colon_pos = lexer.first_colon_pos if ":" in word else None
-
-        return prequote, prefix, suffix, words, first_colon_pos
+        return prequote, prefix, suffix, words, lexer.last_wordbreak_pos
 
     while True:
         try:
@@ -94,19 +106,34 @@ def split_line(line, point=None):
             if lexer.instream.tell() >= point:
                 return split_word(lexer.token)
             else:
-                raise ArgcompleteException("Unexpected internal state. Please report this bug at https://github.com/kislyuk/argcomplete/issues.")  # noqa
+                msg = (
+                    "Unexpected internal state. "
+                    "Please report this bug at https://github.com/kislyuk/argcomplete/issues."
+                )
+                raise ArgcompleteException(msg)
+
 
 def default_validator(completion, prefix):
     return completion.startswith(prefix)
 
+
 class CompletionFinder(object):
     """
     Inherit from this class if you wish to override any of the stages below. Otherwise, use
     ``argcomplete.autocomplete()`` directly (it's a convenience instance of this class). It has the same signature as
     :meth:`CompletionFinder.__call__()`.
     """
-    def __init__(self, argument_parser=None, always_complete_options=True, exclude=None, validator=None,
-                 print_suppressed=False, default_completer=FilesCompleter()):
+
+    def __init__(
+        self,
+        argument_parser=None,
+        always_complete_options=True,
+        exclude=None,
+        validator=None,
+        print_suppressed=False,
+        default_completer=FilesCompleter(),
+        append_space=None,
+    ):
         self._parser = argument_parser
         self.always_complete_options = always_complete_options
         self.exclude = exclude
@@ -117,16 +144,32 @@ class CompletionFinder(object):
         self.completing = False
         self._display_completions = {}
         self.default_completer = default_completer
+        if append_space is None:
+            append_space = os.environ.get("_ARGCOMPLETE_SUPPRESS_SPACE") != "1"
+        self.append_space = append_space
 
-    def __call__(self, argument_parser, always_complete_options=True, exit_method=os._exit, output_stream=None,
-                 exclude=None, validator=None, print_suppressed=False):
+    def __call__(
+        self,
+        argument_parser,
+        always_complete_options=True,
+        exit_method=os._exit,
+        output_stream=None,
+        exclude=None,
+        validator=None,
+        print_suppressed=False,
+        append_space=None,
+        default_completer=FilesCompleter(),
+    ):
         """
         :param argument_parser: The argument parser to autocomplete on
         :type argument_parser: :class:`argparse.ArgumentParser`
         :param always_complete_options:
-            Whether or not to autocomplete options even if an option string opening character (normally ``-``) has not
-            been entered
-        :type always_complete_options: boolean
+            Controls the autocompletion of option strings if an option string opening character (normally ``-``) has not
+            been entered. If ``True`` (default), both short (``-x``) and long (``--x``) option strings will be
+            suggested. If ``False``, no option strings will be suggested. If ``long``, long options and short options
+            with no long variant will be suggested. If ``short``, short options and long options with no short variant
+            will be suggested.
+        :type always_complete_options: boolean or string
         :param exit_method:
             Method used to stop the program after printing completions. Defaults to :meth:`os._exit`. If you want to
             perform a normal exit that calls exit handlers, use :meth:`sys.exit`.
@@ -140,6 +183,9 @@ class CompletionFinder(object):
         :param print_suppressed:
             Whether or not to autocomplete options that have the ``help=argparse.SUPPRESS`` keyword argument set.
         :type print_suppressed: boolean
+        :param append_space:
+            Whether to append a space to unique matches. The default is ``True``.
+        :type append_space: boolean
 
         .. note::
             If you are not subclassing CompletionFinder to override its behaviors,
@@ -151,7 +197,15 @@ class CompletionFinder(object):
         added to argcomplete.safe_actions, if their values are wanted in the ``parsed_args`` completer argument, or
         their execution is otherwise desirable.
         """
-        self.__init__(argument_parser, always_complete_options, exclude, validator, print_suppressed)
+        self.__init__(
+            argument_parser,
+            always_complete_options=always_complete_options,
+            exclude=exclude,
+            validator=validator,
+            print_suppressed=print_suppressed,
+            append_space=append_space,
+            default_completer=default_completer,
+        )
 
         if "_ARGCOMPLETE" not in os.environ:
             # not an argument completion invocation
@@ -160,13 +214,20 @@ class CompletionFinder(object):
         global debug_stream
         try:
             debug_stream = os.fdopen(9, "w")
-        except:
+        except Exception:
             debug_stream = sys.stderr
+        debug()
+
+        if output_stream is None:
+            filename = os.environ.get("_ARGCOMPLETE_STDOUT_FILENAME")
+            if filename is not None:
+                debug("Using output file {}".format(filename))
+                output_stream = open(filename, "w")
 
         if output_stream is None:
             try:
-                output_stream = os.fdopen(8, "wb")
-            except:
+                output_stream = os.fdopen(8, "w")
+            except Exception:
                 debug("Unable to open fd 8 for writing, quitting")
                 exit_method(1)
 
@@ -179,44 +240,60 @@ class CompletionFinder(object):
             debug("Invalid value for IFS, quitting [{v}]".format(v=ifs))
             exit_method(1)
 
+        dfs = os.environ.get("_ARGCOMPLETE_DFS")
+        if dfs and len(dfs) != 1:
+            debug("Invalid value for DFS, quitting [{v}]".format(v=dfs))
+            exit_method(1)
+
         comp_line = os.environ["COMP_LINE"]
         comp_point = int(os.environ["COMP_POINT"])
 
-        # Adjust comp_point for wide chars
-        if USING_PYTHON2:
-            comp_point = len(comp_line[:comp_point].decode(sys_encoding))
-        else:
-            comp_point = len(comp_line.encode(sys_encoding)[:comp_point].decode(sys_encoding))
+        cword_prequote, cword_prefix, cword_suffix, comp_words, last_wordbreak_pos = split_line(comp_line, comp_point)
 
-        comp_line = ensure_str(comp_line)
-        cword_prequote, cword_prefix, cword_suffix, comp_words, first_colon_pos = split_line(comp_line, comp_point)
+        # _ARGCOMPLETE is set by the shell script to tell us where comp_words
+        # should start, based on what we're completing.
+        # 1: <script> [args]
+        # 2: python <script> [args]
+        # 3: python -m <module> [args]
+        start = int(os.environ["_ARGCOMPLETE"]) - 1
+        comp_words = comp_words[start:]
 
-        if os.environ["_ARGCOMPLETE"] == "2":
-            # Shell hook recognized the first word as the interpreter; discard it
-            comp_words.pop(0)
+        if cword_prefix and cword_prefix[0] in self._parser.prefix_chars and "=" in cword_prefix:
+            # Special case for when the current word is "--optional=PARTIAL_VALUE". Give the optional to the parser.
+            comp_words.append(cword_prefix.split("=", 1)[0])
 
-        debug("\nLINE: '{l}'\nPREQUOTE: '{pq}'\nPREFIX: '{p}'".format(l=comp_line, pq=cword_prequote, p=cword_prefix),
-              "\nSUFFIX: '{s}'".format(s=cword_suffix),
-              "\nWORDS:", comp_words)
+        debug(
+            "\nLINE: {!r}".format(comp_line),
+            "\nPOINT: {!r}".format(comp_point),
+            "\nPREQUOTE: {!r}".format(cword_prequote),
+            "\nPREFIX: {!r}".format(cword_prefix),
+            "\nSUFFIX: {!r}".format(cword_suffix),
+            "\nWORDS:",
+            comp_words,
+        )
 
-        completions = self._get_completions(comp_words, cword_prefix, cword_prequote, first_colon_pos)
+        completions = self._get_completions(comp_words, cword_prefix, cword_prequote, last_wordbreak_pos)
+
+        if dfs:
+            display_completions = {
+                key_part: value.replace(ifs, " ") if value else ""
+                for key, value in self._display_completions.items()
+                for key_part in key
+            }
+            completions = [dfs.join((key, display_completions.get(key) or "")) for key in completions]
 
         debug("\nReturning completions:", completions)
-        output_stream.write(ifs.join(completions).encode(sys_encoding))
+        output_stream.write(ifs.join(completions))
         output_stream.flush()
         debug_stream.flush()
         exit_method(0)
 
-    def _get_completions(self, comp_words, cword_prefix, cword_prequote, first_colon_pos):
+    def _get_completions(self, comp_words, cword_prefix, cword_prequote, last_wordbreak_pos):
         active_parsers = self._patch_argument_parser()
 
         parsed_args = argparse.Namespace()
         self.completing = True
 
-        if USING_PYTHON2:
-            # Python 2 argparse only properly works with byte strings.
-            comp_words = [ensure_bytes(word) for word in comp_words]
-
         try:
             debug("invoking parser with", comp_words[1:])
             with mute_stderr():
@@ -227,11 +304,12 @@ class CompletionFinder(object):
 
         self.completing = False
 
-        # key: complete word, value: description.
+        if "--" in comp_words:
+            self.always_complete_options = False
 
         completions = self.collect_completions(active_parsers, parsed_args, cword_prefix, debug)
         completions = self.filter_completions(completions)
-        completions = self.quote_completions(completions, cword_prequote, first_colon_pos)
+        completions = self.quote_completions(completions, cword_prequote, last_wordbreak_pos)
         return completions
 
     def _patch_argument_parser(self):
@@ -254,8 +332,7 @@ class CompletionFinder(object):
                 return
 
             classname = "MonkeyPatchedIntrospectiveArgumentParser"
-            if USING_PYTHON2:
-                classname = bytes(classname)
+
             parser.__class__ = type(classname, (IntrospectiveArgumentParser, parser.__class__), {})
 
             for action in parser._actions:
@@ -295,42 +372,70 @@ class CompletionFinder(object):
         return self.active_parsers
 
     def _get_subparser_completions(self, parser, cword_prefix):
-        def filter_aliases(metavar, dest, prefix):
-            if not metavar:
-                return dest if dest and dest.startswith(prefix) else ""
+        def filter_aliases(aliases, prefix):
+            return tuple(x for x in aliases if x.startswith(prefix))
 
-            # metavar combines dest and aliases with ",".
-            a = metavar.replace(",", "").split()
-            return " ".join(x for x in a if x.startswith(prefix))
+        aliases_by_parser = {}
+        for key in parser.choices.keys():
+            p = parser.choices[key]
+            aliases_by_parser.setdefault(p, []).append(key)
 
         for action in parser._get_subactions():
-            subcmd_with_aliases = filter_aliases(action.metavar, action.dest, cword_prefix)
+            subcmd_with_aliases = filter_aliases(aliases_by_parser[parser.choices[action.dest]], cword_prefix)
             if subcmd_with_aliases:
                 self._display_completions[subcmd_with_aliases] = action.help
 
         completions = [subcmd for subcmd in parser.choices.keys() if subcmd.startswith(cword_prefix)]
         return completions
 
+    def _include_options(self, action, cword_prefix):
+        if len(cword_prefix) > 0 or self.always_complete_options is True:
+            return [opt for opt in action.option_strings if opt.startswith(cword_prefix)]
+        long_opts = [opt for opt in action.option_strings if len(opt) > 2]
+        short_opts = [opt for opt in action.option_strings if len(opt) <= 2]
+        if self.always_complete_options == "long":
+            return long_opts if long_opts else short_opts
+        elif self.always_complete_options == "short":
+            return short_opts if short_opts else long_opts
+        return []
+
     def _get_option_completions(self, parser, cword_prefix):
         self._display_completions.update(
-            [[" ".join(ensure_str(x) for x in action.option_strings if ensure_str(x).startswith(cword_prefix)), action.help]  # noqa
-             for action in parser._actions
-             if action.option_strings])
+            [
+                [tuple(x for x in action.option_strings if x.startswith(cword_prefix)), action.help]
+                for action in parser._actions
+                if action.option_strings
+            ]
+        )
 
         option_completions = []
         for action in parser._actions:
-            if action.help == argparse.SUPPRESS and not self.print_suppressed:
+            if not self.print_suppressed:
+                completer = getattr(action, "completer", None)
+                if isinstance(completer, SuppressCompleter) and completer.suppress():
+                    continue
+                if action.help == argparse.SUPPRESS:
+                    continue
+            if not self._action_allowed(action, parser):
                 continue
             if not isinstance(action, argparse._SubParsersAction):
-                for option in action.option_strings:
-                    if ensure_str(option).startswith(cword_prefix):
-                        option_completions.append(ensure_str(option))
+                option_completions += self._include_options(action, cword_prefix)
         return option_completions
 
+    @staticmethod
+    def _action_allowed(action, parser):
+        # Logic adapted from take_action in ArgumentParser._parse_known_args
+        # (members are saved by my_argparse.IntrospectiveArgumentParser)
+        for conflict_action in parser._action_conflicts.get(action, []):
+            if conflict_action in parser._seen_non_default_actions:
+                return False
+        return True
+
     def _complete_active_option(self, parser, next_positional, cword_prefix, parsed_args, completions):
         debug("Active actions (L={l}): {a}".format(l=len(parser.active_actions), a=parser.active_actions))
 
         isoptional = cword_prefix and cword_prefix[0] in parser.prefix_chars
+        optional_prefix = ""
         greedy_actions = [x for x in parser.active_actions if action_is_greedy(x, isoptional)]
         if greedy_actions:
             assert len(greedy_actions) == 1, "expect at most 1 greedy action"
@@ -340,8 +445,14 @@ class CompletionFinder(object):
             self._display_completions = {}
             completions = []
         elif isoptional:
-            # Only run completers if current word does not start with - (is not an optional)
-            return completions
+            if "=" in cword_prefix:
+                # Special case for when the current word is "--optional=PARTIAL_VALUE".
+                # The completer runs on PARTIAL_VALUE. The prefix is added back to the completions
+                # (and chopped back off later in quote_completions() by the COMP_WORDBREAKS logic).
+                optional_prefix, _, cword_prefix = cword_prefix.partition("=")
+            else:
+                # Only run completers if current word does not start with - (is not an optional)
+                return completions
 
         complete_remaining_positionals = False
         # Use the single greedy action (if there is one) or all active actions.
@@ -370,19 +481,26 @@ class CompletionFinder(object):
                     completer = self.default_completer
 
             if completer:
+                if isinstance(completer, SuppressCompleter) and completer.suppress():
+                    continue
+
                 if callable(completer):
-                    completions_from_callable = [c for c in completer(
-                        prefix=cword_prefix, action=active_action, parsed_args=parsed_args)
-                        if self.validator(c, cword_prefix)]
+                    completions_from_callable = [
+                        c
+                        for c in completer(
+                            prefix=cword_prefix, action=active_action, parser=parser, parsed_args=parsed_args
+                        )
+                        if self.validator(c, cword_prefix)
+                    ]
 
                     if completions_from_callable:
                         completions += completions_from_callable
                         if isinstance(completer, completers.ChoicesCompleter):
                             self._display_completions.update(
-                                [[x, active_action.help] for x in completions_from_callable])
+                                [[(x,), active_action.help] for x in completions_from_callable]
+                            )
                         else:
-                            self._display_completions.update(
-                                [[x, ""] for x in completions_from_callable])
+                            self._display_completions.update([[(x,), ""] for x in completions_from_callable])
                 else:
                     debug("Completer is not callable, trying the readline completer protocol instead")
                     for i in range(9999):
@@ -390,8 +508,10 @@ class CompletionFinder(object):
                         if next_completion is None:
                             break
                         if self.validator(next_completion, cword_prefix):
-                            self._display_completions.update({next_completion: ""})
+                            self._display_completions.update({(next_completion,): ""})
                             completions.append(next_completion)
+                if optional_prefix:
+                    completions = [optional_prefix + "=" + completion for completion in completions]
                 debug("Completions:", completions)
         return completions
 
@@ -417,8 +537,9 @@ class CompletionFinder(object):
         if isinstance(next_positional, argparse._SubParsersAction):
             completions += self._get_subparser_completions(next_positional, cword_prefix)
 
-        completions = self._complete_active_option(active_parser, next_positional, cword_prefix, parsed_args,
-                                                   completions)
+        completions = self._complete_active_option(
+            active_parser, next_positional, cword_prefix, parsed_args, completions
+        )
         debug("active options:", completions)
         debug("display completions:", self._display_completions)
 
@@ -455,18 +576,13 @@ class CompletionFinder(object):
 
         This method is exposed for overriding in subclasses; there is no need to use it directly.
         """
-        # On Python 2, we have to make sure all completions are unicode objects before we continue and output them.
-        # Otherwise, because python disobeys the system locale encoding and uses ascii as the default encoding, it will
-        # try to implicitly decode string objects using ascii, and fail.
-        completions = [ensure_str(c) for c in completions]
-
         # De-duplicate completions and remove excluded ones
         if self.exclude is None:
             self.exclude = set()
         seen = set(self.exclude)
         return [c for c in completions if c not in seen and not seen.add(c)]
 
-    def quote_completions(self, completions, cword_prequote, first_colon_pos):
+    def quote_completions(self, completions, cword_prequote, last_wordbreak_pos):
         """
         If the word under the cursor started with a quote (as indicated by a nonempty ``cword_prequote``), escapes
         occurrences of that quote character in the completions, and adds the quote to the beginning of each completion.
@@ -478,45 +594,46 @@ class CompletionFinder(object):
 
         This method is exposed for overriding in subclasses; there is no need to use it directly.
         """
-        comp_wordbreaks = ensure_str(os.environ.get("_ARGCOMPLETE_COMP_WORDBREAKS",
-                                                    os.environ.get("COMP_WORDBREAKS",
-                                                                   " \t\"'@><=;|&(:.")))
-
-        punctuation_chars = "();<>|&!`"
-        for char in punctuation_chars:
-            if char not in comp_wordbreaks:
-                comp_wordbreaks += char
-
-        # If the word under the cursor was quoted, escape the quote char and add the leading quote back in.
-        # Otherwise, escape all COMP_WORDBREAKS chars.
+        special_chars = "\\"
+        # If the word under the cursor was quoted, escape the quote char.
+        # Otherwise, escape all special characters and specially handle all COMP_WORDBREAKS chars.
         if cword_prequote == "":
-            # Bash mangles completions which contain colons if COMP_WORDBREAKS contains a colon.
-            # This workaround has the same effect as __ltrim_colon_completions in bash_completion.
-            if ":" in comp_wordbreaks and first_colon_pos:
-                completions = [c[first_colon_pos + 1:] for c in completions]
+            # Bash mangles completions which contain characters in COMP_WORDBREAKS.
+            # This workaround has the same effect as __ltrim_colon_completions in bash_completion
+            # (extended to characters other than the colon).
+            if last_wordbreak_pos:
+                completions = [c[last_wordbreak_pos + 1 :] for c in completions]
+            special_chars += "();<>|&!`$* \t\n\"'"
+        elif cword_prequote == '"':
+            special_chars += '"`$!'
 
-            for wordbreak_char in comp_wordbreaks:
-                completions = [c.replace(wordbreak_char, "\\" + wordbreak_char) for c in completions]
-        else:
-            if cword_prequote == '"':
-                for char in "`$!":
-                    completions = [c.replace(char, "\\" + char) for c in completions]
-            completions = [cword_prequote + c.replace(cword_prequote, "\\" + cword_prequote) for c in completions]
+        if os.environ.get("_ARGCOMPLETE_SHELL") in ("tcsh", "fish"):
+            # tcsh and fish escapes special characters itself.
+            special_chars = ""
+        elif cword_prequote == "'":
+            # Nothing can be escaped in single quotes, so we need to close
+            # the string, escape the single quote, then open a new string.
+            special_chars = ""
+            completions = [c.replace("'", r"'\''") for c in completions]
 
-        # Note: similar functionality in bash is turned off by supplying the "-o nospace" option to complete.
-        # We can't use that functionality because bash is not smart enough to recognize continuation characters (/) for
-        # which no space should be added.
-        continuation_chars = "=/:"
-        if len(completions) == 1 and completions[0][-1] not in continuation_chars:
-            if cword_prequote == "" and not completions[0].endswith(" "):
-                completions[0] += " "
+        for char in special_chars:
+            completions = [c.replace(char, "\\" + char) for c in completions]
+
+        if self.append_space:
+            # Similar functionality in bash was previously turned off by supplying the "-o nospace" option to complete.
+            # Now it is conditionally disabled using "compopt -o nospace" if the match ends in a continuation character.
+            # This code is retained for environments where this isn't done natively.
+            continuation_chars = "=/:"
+            if len(completions) == 1 and completions[0][-1] not in continuation_chars:
+                if cword_prequote == "":
+                    completions[0] += " "
 
         return completions
 
     def rl_complete(self, text, state):
         """
         Alternate entry point for using the argcomplete completer in a readline-based REPL. See also
-        `rlcompleter <https://docs.python.org/2/library/rlcompleter.html#completer-objects>`_.
+        `rlcompleter <https://docs.python.org/3/library/rlcompleter.html#completer-objects>`_.
         Usage:
 
         .. code-block:: python
@@ -529,14 +646,12 @@ class CompletionFinder(object):
             readline.set_completer(completer.rl_complete)
             readline.parse_and_bind("tab: complete")
             result = input("prompt> ")
-
-        (Use ``raw_input`` instead of ``input`` on Python 2, or use `eight <https://github.com/kislyuk/eight>`_).
         """
         if state == 0:
             cword_prequote, cword_prefix, cword_suffix, comp_words, first_colon_pos = split_line(text)
             comp_words.insert(0, sys.argv[0])
             matches = self._get_completions(comp_words, cword_prefix, cword_prequote, first_colon_pos)
-            self._rl_matches = [text + match[len(cword_prefix):] for match in matches]
+            self._rl_matches = [text + match[len(cword_prefix) :] for match in matches]
 
         if state < len(self._rl_matches):
             return self._rl_matches[state]
@@ -574,14 +689,35 @@ class CompletionFinder(object):
             readline.set_completion_display_matches_hook(display_completions)
 
         """
-        return self._display_completions
+        return {" ".join(k): v for k, v in self._display_completions.items()}
+
+
+class ExclusiveCompletionFinder(CompletionFinder):
+    @staticmethod
+    def _action_allowed(action, parser):
+        if not CompletionFinder._action_allowed(action, parser):
+            return False
+
+        append_classes = (argparse._AppendAction, argparse._AppendConstAction)
+        if action._orig_class in append_classes:
+            return True
+
+        if action not in parser._seen_non_default_actions:
+            return True
+
+        return False
+
 
 autocomplete = CompletionFinder()
 autocomplete.__doc__ = """ Use this to access argcomplete. See :meth:`argcomplete.CompletionFinder.__call__()`. """
 
+
 def warn(*args):
     """
     Prints **args** to standard error when running completions. This will interrupt the user's command line interaction;
     use it to indicate an error condition that is preventing your completer from working.
     """
-    print("\n", file=debug_stream, *args)
+    # Don't be tempted to use `print("\n",..., *args)`,
+    # as that will indent **args** by one space character
+    print(file=debug_stream)
+    print(file=debug_stream, *args)
